/*
zuisti's Scroll16c-01.c, but modified (by zuisti) for a 80 columns matrix

- in reality use row-side Common Anode matrices and drivers !!!!!!!!!!!!!

- PIC16F877A,8 MHz

- 16 pcs of 7*5 or 10 pcs of 8*8 matrices (in this case the 8th row is unused
                                           because of the old bitmap table)

- the timer2 setting is also recalculated (new time measurements are made in Proteus)
  (now the refresh rate is 71.4 Hz at 8 Mhz)
*/

#define SCROLLTIME 5 // loop count value, the smaller is faster (fastest: 1)

// hw defines:
#define COMMON_CATHODE       //now: common cathode rows, common anode columns !!!

#define DATA_PIN  RB3_bit //PORTB.3, the hc595 data

//Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
#define StorePulse RA3_bit = 1; RA3_bit = 0

//clock pin: PORTB.0  (RB0), positive clock for shift in
#define ClockPulse RB0_bit = 1; RB0_bit = 0

#define SCROLL_PIN RB4_bit   // a jumper for (re)start (open) / stop (closed) scrolling

//Line decoder inputs: PORTA<2:0> ;RA0 = LSB
#define DisableLines PORTA = 0b0111  //= 7, all 1 (the ls138 outputs are active low)

#include "bitmap7.h" // old ASCII font bitmap table, 7 bytes/char (!!)

const char message[] = "Helped (modified) by zuisti, good luck !!      ";

//global arrays
char pattbuf[7];   // holds the actual char's line patterns
char linebuf[10*7]; // 10 bytes for the first (uppermost) line (row), 10 bytes for the 2nd, ...

void HW_Init(void) {
    ADCON1 = 0x07;      // all analog off
    CMCON = 0b00000111; // switch off comparators
    DisableLines;       // PORTA = 7, also store pin is zeroed
    PORTB = 0;          // also data and clock pins are zeroed
    TRISA = 0;          // PORTA pins are outputs
    TRISB = 0xF6;       // only PORTB.0, 3 are outputs
    NOT_RBPU_bit = 0;   // portb pullups on

/*
init timer2 for line timing (same row brightness)
    one row display = 0.62 ms (without timer2 waiting)
    but with scroll (worst case 1.38 ms) = sum 2 ms (value for timer2)

    Prescaler=1:1; PostScaler=1:16; PR2 = 248; Period = 2.0 ms  (8 MHz)
    8 MHz: 2 ms/line, 14 ms the whole 7 lines loop (71.4 Hz refresh rate)
*/
    PR2 = 248;           // Timer2 Match value for 2 ms (8 MHz)
    T2CON = 0b01111100;  // enable timer2 with 1:16 (!!) postscaler, 1:1 prescaler
    TMR2IF_bit = 0;      // clear IT flag
    PIE1 = 2;            // only TMR2IE_bit = 1: enable Timer2 IT

    INTCON = 0xC0;       // Enable global (GIE_bit) and periferial (PEIE_bit) ITs
}

void interrupt(void) {
    DisableLines;   // PORTA = 7
    TMR2ON_bit = 0; // disable timer2
    TMR2IF_bit = 0; // clear IT flag
}

void get_patterns(char ch) {
     unsigned bitmap_idx;
     char rowcnt;

     ch -= 32; // the space is the first char in the bitmap table
     //calculate the bitmap table offset of the first pattern of the char ch
     bitmap_idx = 8*ch - ch; // = 7*ch, but faster

     rowcnt = 0; //we have 7 patterns (0...6)
     do {
        ch = bitmap[bitmap_idx + rowcnt]; // read pattern from the bitmap table
        ch <<= 2; // drop unused two bits (7,6)
#ifdef COMMON_CATHODE
        ch = ~ch; // inverted (0<>1)
#endif
        pattbuf[rowcnt] = ch; // and store it
     } while (++rowcnt != 7);
}

char nextchar(void) {
     char ch;
     static unsigned msg_idx = 0; // 16bit var, a very long msg is also possible

     for (;;) {
         if (ch = message[msg_idx++])
            break;    // not 0 (end sign)
         msg_idx = 0; // else back to begin
     }
     return ch;
}

void shift_in_a_column(void) {
     char pat_idx, buf_idx, ch, rowcnt, bytecnt, carry_over;

     pat_idx = 6;  // last (7th, lowermost) pattern
     buf_idx = 69; // last (7th, lowermost) line's last (10th) byte

     rowcnt = 7;
     do {
        // unfortunately (in MikroC) no rotate with carry

        // pattern rotate left with carry
        ch = pattbuf[pat_idx];
        carry_over = ch & 0x80;
        ch <<= 1; // a 0 shifted in from right but not used
        pattbuf[pat_idx--] = ch;

        // rotate a whole line (10 bytes) left with carry
        bytecnt = 10;
        do {
           ch = linebuf[buf_idx];
           ch <<= 1;
           if (carry_over) ch.B0 = 1;
           linebuf[buf_idx--] = ch;
           carry_over = ch & 0x80;
        } while (--bytecnt);
     } while (--rowcnt);
}

void display(void) {
   char ch, buf_idx, bytecnt, bitcnt, rowcnt;

   buf_idx = 0; //from begin of buffer
   // loop until rowcnt = 7
   rowcnt = 0;
   do { // 617 us / row (8 Mhz)
      bytecnt = 10;
      do {
         ch = linebuf[buf_idx++];
         bitcnt = 8;
         do {
            DATA_PIN = ch.B7;
            ch <<= 1;
            ClockPulse;
         } while (--bitcnt);
      }while (--bytecnt);

      //timed display (same row brightness)
      do {} while (TMR2ON_bit); //wait for timer2 IT (stop timer2 and all lines off)

      TMR2ON_bit = 1; //restart timer2, also a time gap to switch off drivers

      StorePulse;
      PORTA = rowcnt; // switch on the actual line driver
   } while (++rowcnt != 7);
}

void clear_linebuf(void) {
  char cnt = 70; // 10*7
  do {
#ifdef COMMON_CATHODE
     linebuf[cnt-1] = 0;
#else
     linebuf[cnt-1] = 255;
#endif
  } while (--cnt);
}

void main(void) {
  char width, scrollcnt, ch;

  HW_Init();
  clear_linebuf();
  //for prompt scroll:
  width = 0;
  scrollcnt = 1;

  for (;;) {
    if (SCROLL_PIN && --scrollcnt == 0) {
       // max sum scroll time (if width = 0) = 1.38 ms  (8 MHz)
       scrollcnt = SCROLLTIME;
       if (!width) {
          ch = nextchar();  //26 us (8 MHz)
          width = 6;
          if (ch == 32) width = 3;
          get_patterns(ch); //218 us  (8 MHz)
       }
       --width;
       shift_in_a_column(); //1122 us !!!!  (8 MHz)
    }
    display(); //appr 0.62 ms /row, without timer2 waiting !  (8 MHz)
               //so max. row-time = 1.38 + 0.62 = 2 ms
  }
}
//------ end